@@ -4,8 +4,8 @@
 # This file is part of LEPL.
 # 
 #     LEPL is free software: you can redistribute it and/or modify
-#     it under the terms of the GNU Lesser General Public License as published by
-#     the Free Software Foundation, either version 3 of the License, or
+#     it under the terms of the GNU Lesser General Public License as published 
+#     by the Free Software Foundation, either version 3 of the License, or
 #     (at your option) any later version.
 # 
 #     LEPL is distributed in the hope that it will be useful,
@@ -20,82 +20,21 @@
 Support for operator syntactic sugar (and operator redefinition).
 '''
 
-from threading import local
 from abc import ABCMeta
 
+from lepl.context import Namespace, NamespaceMixin, Scope
 from lepl.support import open_stop
 
 
+# pylint: disable-msg=C0103, W0105
 # Python 2.6
 #class Matcher(metaclass=ABCMeta):
 Matcher = ABCMeta('Matcher', (object, ), {})
-'''ABC used to identify matchers.'''
+'''
+ABC used to identify matchers.  
 
-
-class Namespace(local):
-    '''
-    A store for operator definitions.
-    
-    This subclasses threading.local so each thread effectively has its own 
-    instance (see test).
-    '''
-    
-    def __init__(self, base={}):
-        super(Namespace, self).__init__()
-        self.__stack = [base]
-        
-    def push(self, extra={}):
-        '''
-        Add further names, savign the current state to the stack.
-        '''
-        self.__stack.append(dict(self.current()))
-        for name in extra:
-            self.set_opt(name, extra[name])
-        
-    def pop(self):
-        '''
-        Return the previous state from the stack.
-        '''
-        self.__stack.pop(-1)
-        
-    def __enter__(self):
-        '''
-        Allow use within a with context by duplicating the current state
-        and saving to the stack.  Returns self to allow manipulation via set.
-        '''
-        self.push()
-        return self
-        
-    def __exit__(self, *args):
-        '''
-        Restore the previous state from the stack on leaving the context.
-        '''
-        self.pop()
-        
-    def current(self):
-        '''
-        The current state (a map from names to operator implementations).
-        '''
-        return self.__stack[-1]
-    
-    def set(self, name, value):
-        '''
-        Set a value.
-        '''
-        self.current()[name] = value
-        
-    def set_opt(self, name, value):
-        '''
-        Set a value if it is not None.
-        '''
-        if value != None:
-            self.set(name, value)
-        
-    def __getitem__(self, name):
-        '''
-        Get a value.
-        '''
-        return self.current()[name]
+Note that graph traversal assumes subclasses are hashable and iterable.
+'''
 
 
 class DefaultNamespace(Namespace):
@@ -105,8 +44,8 @@ class DefaultNamespace(Namespace):
     
     def __init__(self):
         # Handle circular dependencies
-        from lepl.matchers import And, Space, Add, Or, Apply, Drop, KApply, \
-            Repeat, raise_error, First, Map
+        from lepl.matchers import And, Or, raise_error, First
+        from lepl.functions import Space, Add, Apply, KApply, Drop, Repeat, Map
         super(DefaultNamespace, self).__init__({
             SPACE_OPT: lambda a, b: And(a, Space()[0:,...], b),
             SPACE_REQ: lambda a, b: And(a, Space()[1:,...], b),
@@ -114,8 +53,8 @@ class DefaultNamespace(Namespace):
             AND:       And,
             OR:        Or,
             APPLY:     Apply,
+            APPLY_RAW: lambda a, b: Apply(a, b, raw=True),
             NOT:       Drop,
-            ARGS:      lambda a, b: Apply(a, b, args=True),
             KARGS:     KApply,
             RAISE:     lambda a, b: KApply(a, raise_error(b)),
             REPEAT:    Repeat,
@@ -124,24 +63,11 @@ class DefaultNamespace(Namespace):
         })
     
 
-__GLOBAL = None
+OPERATORS = 'operators'
 '''
-The global singleton.
+The name used to retrieve operators definitions.
 '''
 
-def Global():
-    '''
-    Global (per-thread) binding from operator name to implementation.
-    
-    This provides an interface to `__GLOBAL`.
-    '''
-    # Delay creation to handle circular dependencies.
-    global __GLOBAL
-    if __GLOBAL is None:
-        __GLOBAL = DefaultNamespace()
-    return __GLOBAL
-
-
 SPACE_OPT = '/'
 '''Name for / operator.'''
 SPACE_REQ = '//'
@@ -154,10 +80,10 @@ OR = '|'
 '''Name for | operator.'''
 APPLY = '>'
 '''Name for > operator.'''
+APPLY_RAW = '>='
+'''Name for >= operator.'''
 NOT = '~'
 '''Name for ~ operator.'''
-ARGS = '*'
-'''Name for * operator.'''
 KARGS = '**'
 '''Name for ** operator.'''
 RAISE = '^'
@@ -170,53 +96,116 @@ MAP = '>>'
 '''Name for >> operator.'''
 
 
-class Override(object):
+class Override(Scope):
     '''
-    Allow an operator to be redefined within a with context.
+    Allow an operator to be redefined within a with context.  Uses the 
+    OPERATORS namespace.
     '''
 
     def __init__(self, space_opt=None, space_req=None, repeat=None,
                   add=None, and_=None, or_=None, not_=None, 
-                  apply=None, args=None, kargs=None, raise_=None,
-                  first=None, map=None):
-        self.__frame ={SPACE_OPT: space_opt, SPACE_REQ: space_req,
-                       REPEAT: repeat, ADD: add, AND: and_, OR: or_, 
-                       NOT: not_, APPLY: apply, ARGS: args, KARGS: kargs, 
-                       RAISE: raise_, FIRST: first, MAP: map}
-        
-    def __enter__(self):
-        '''
-        On entering the context, add the new definitions.
-        '''
-        Global().push(self.__frame)
-        
-    def __exit__(self, *args):
-        '''
-        On leaving the context, return to previous definition.
-        '''
-        Global().pop()
+                  apply_=None, apply_raw=None, kargs=None, 
+                  raise_=None, first=None, map_=None):
+        super(Override, self).__init__(OPERATORS, DefaultNamespace,
+            {SPACE_OPT: space_opt, SPACE_REQ: space_req,
+             REPEAT: repeat, ADD: add, AND: and_, OR: or_, 
+             NOT: not_, APPLY: apply_, APPLY_RAW: apply_raw,
+             KARGS: kargs, RAISE: raise_, FIRST: first, MAP: map_})
 
 
-class Separator(Override):
+class _BaseSeparator(Override):
     '''
-    Redefine ``[]`` and ``&`` to include the given matcher as a separator 
-    (so it will be used between list items and between matchers separated by the & 
-    operator)
+    Support class for `Separator` and similar classes.
+    
+    Uses the OPERATORS namespace.
     '''
     
     def __init__(self, separator):
         '''
-        If the separator is a string it is coerced to `lepl.matchers.Regexp()`.
+        If the separator is a string it is coerced to `Regexp()`; if None
+        then any previous defined separator is effectively removed.
         '''
         # Handle circular dependencies
-        from lepl.matchers import Regexp, And, Repeat, coerce
-        separator = coerce(separator, Regexp)
-        and_ = lambda a, b: And(a, separator, b)
+        from lepl.functions import Repeat
+        from lepl.matchers import Regexp, And, coerce_
+        if separator is None:
+            and_ = And
+            repeat = Repeat
+        else:
+            separator = coerce_(separator, Regexp)
+            (and_, repeat) = self._replacements(separator)
+        super(_BaseSeparator, self).__init__(and_=and_, repeat=repeat)
+
+    def _replacements(self, _separator):
+        '''
+        Sub-classes should return (And, Repeat)
+        '''
+        raise Exception('Unimplemented')
+    
+    def _repeat(self, separator):
+        '''
+        A simple Repeat with separator.
+        '''
+        from lepl.functions import Repeat
+        from lepl.matchers import And
         def repeat(m, st=0, sp=None, d=0, s=None, a=False):
-            if s is None: s = separator
+            '''
+            Wrap `Repeat` to adapt the separator.
+            '''
+            if s is None:
+                s = separator
+            elif not a:
+                s = And(separator, s, separator)
             return Repeat(m, st, sp, d, s, a)
-        super(Separator, self).__init__(and_=and_, repeat=repeat)
+        return repeat
+    
+
+class Separator(_BaseSeparator):
+    '''
+    Redefine ``[]`` and ``&`` to include the given matcher as a separator 
+    (so it will be used between list items and between matchers separated by the & 
+    operator)
+    
+    Uses the OPERATORS namespace.
+    '''
+    
+    def _replacements(self, separator):
+        '''
+        Require the separator on each `And`.
+        '''
+        # Handle circular dependencies
+        from lepl.matchers import And
+        return (lambda a, b: And(a, separator, b),
+                self._repeat(separator))
         
+
+class SmartSeparator1(_BaseSeparator):
+    '''
+    Similar to `Separator`, but tried to be clever about whether the 
+    separator is needed.  It replaces `&` with a matcher that only uses 
+    the separator if the second sub-matcher consumes some input.
+    
+    Uses the OPERATORS namespace.
+    
+    See also `SmartSeparator2`, which is less general, but more efficient.
+    '''
+    
+    def _replacements(self, separator):
+        '''
+        Require the separator on each `And`.
+        '''
+        # Handle circular dependencies
+        from lepl.matchers import Consumer, And, Or
+        def and_(a, b):
+            '''
+            Add space only in the case when both consume something.
+            '''
+            return Or(And(Consumer(a), separator, Consumer(b)),
+                      And(Consumer(a), Consumer(b, False)),
+                      And(Consumer(a, False), Consumer(b)),
+                      And(Consumer(a, False), Consumer(b, False)))
+        return (and_, self._repeat(separator))
+   
         
 GREEDY = 'g'
 '''Flag (splice increment) for inefficient, guaranteed greedy matching.'''
@@ -228,13 +217,13 @@ BREADTH_FIRST = 'b'
 '''Flag (splice increment) for efficient, quasi-non-greedy, matching.'''
 
 
-class OperatorMixin(object):
+class OperatorMixin(NamespaceMixin):
     '''
     Define the operators used to combine elements in a grammar specification.
     '''
 
-    def __init__(self):
-        super(OperatorMixin, self).__init__()
+    def __init__(self, name, namespace):
+        super(OperatorMixin, self).__init__(name, namespace)
         
     def __add__(self, other):
         '''
@@ -250,7 +239,7 @@ class OperatorMixin(object):
             match.
         '''
         self.__check(ADD, other, True)
-        return Global()[ADD](self, other)
+        return self._lookup(ADD)(self, other)
 
     def __radd__(self, other):
         '''
@@ -266,14 +255,14 @@ class OperatorMixin(object):
             match.
         '''
         self.__check(ADD, other, True)
-        return Global()[ADD](other, self)
+        return self._lookup(ADD)(other, self)
 
     def __and__(self, other):
         '''
         **self & other** - Append results.
         
         Combine adjacent matchers in sequence.  This is equivalent to 
-        `lepl.matchers.And()`.
+        `And()`.
         
         :Parameters:
         
@@ -282,14 +271,14 @@ class OperatorMixin(object):
             match.
         '''
         self.__check(AND, other, True)
-        return Global()[AND](self, other) 
+        return self._lookup(AND)(self, other) 
         
     def __rand__(self, other):
         '''
         **other & self** - Append results.
         
         Combine adjacent matchers in sequence.  This is equivalent to 
-        `lepl.matchers.And()`.
+        `And()`.
         
         :Parameters:
         
@@ -298,7 +287,7 @@ class OperatorMixin(object):
             match.
         '''
         self.__check(AND, other, True)
-        return Global()[AND](other, self)
+        return self._lookup(AND)(other, self)
     
     def __div__(self, other):
         '''
@@ -326,7 +315,7 @@ class OperatorMixin(object):
             match.
         '''
         self.__check(SPACE_OPT, other, True)
-        return Global()[SPACE_OPT](self, other)
+        return self._lookup(SPACE_OPT)(self, other)
         
     def __rtruediv__(self, other):
         '''
@@ -342,7 +331,7 @@ class OperatorMixin(object):
             match.
         '''
         self.__check(SPACE_OPT, other, True)
-        return Global()[SPACE_OPT](other, self)
+        return self._lookup(SPACE_OPT)(other, self)
         
     def __floordiv__(self, other):
         '''
@@ -358,7 +347,7 @@ class OperatorMixin(object):
             match.
         '''
         self.__check(SPACE_REQ, other, True)
-        return Global()[SPACE_REQ](self, other)
+        return self._lookup(SPACE_REQ)(self, other)
         
     def __rfloordiv__(self, other):
         '''
@@ -374,7 +363,7 @@ class OperatorMixin(object):
             match.
         '''
         self.__check(SPACE_REQ, other, True)
-        return Global()[SPACE_REQ](other, self)
+        return self._lookup(SPACE_REQ)(other, self)
         
     def __or__(self, other):
         '''
@@ -382,7 +371,7 @@ class OperatorMixin(object):
         
         This introduces backtracking.  Matches are tried from left to right
         and successful results returned (one on each "recall").  This is 
-        equivalent to `lepl.matchers.Or()`.
+        equivalent to `Or()`.
         
         :Parameters:
         
@@ -391,7 +380,7 @@ class OperatorMixin(object):
             match.
         '''
         self.__check(OR, other, True)
-        return Global()[OR](self, other) 
+        return self._lookup(OR)(self, other) 
         
     def __ror__(self, other):
         '''
@@ -399,7 +388,7 @@ class OperatorMixin(object):
         
         This introduces backtracking.  Matches are tried from left to right
         and successful results returned (one on each "recall").  This is 
-        equivalent to `lepl.matchers.Or()`.
+        equivalent to `Or()`.
         
         :Parameters:
         
@@ -408,14 +397,14 @@ class OperatorMixin(object):
             match.
         '''
         self.__check(OR, other, True)
-        return Global()[OR](other, self) 
+        return self._lookup(OR)(other, self) 
         
     def __mod__(self, other):
         '''
         **self % other** - Take first match (committed choice).
         
         Matches are tried from left to right and the first successful result
-        is returned.  This is equivalent to `lepl.matchers.First()`.
+        is returned.  This is equivalent to `First()`.
         
         :Parameters:
         
@@ -424,14 +413,14 @@ class OperatorMixin(object):
             match.
         '''
         self.__check(FIRST, other, True)
-        return Global()[FIRST](self, other) 
+        return self._lookup(FIRST)(self, other) 
         
     def __rmod__(self, other):
         '''
         **other % self** - Take first match (committed choice).
         
         Matches are tried from left to right and the first successful result
-        is returned.  This is equivalent to `lepl.matchers.First()`.
+        is returned.  This is equivalent to `First()`.
         
         :Parameters:
         
@@ -440,19 +429,19 @@ class OperatorMixin(object):
             match.
         '''
         self.__check(FIRST, other, True)
-        return Global()[FIRST](other, self) 
+        return self._lookup(FIRST)(other, self) 
         
     def __invert__(self):
         '''
         **~self** - Discard the result.
 
         This generates a matcher that behaves as the original, but returns
-        an empty list. This is equivalent to `lepl.matchers.Drop()`.
+        an empty list. This is equivalent to `Drop()`.
         
-        Note that `lepl.matchers.Lookahead()` overrides this method to have
+        Note that `Lookahead()` overrides this method to have
         different semantics (negative lookahead).
         '''
-        return Global()[NOT](self) 
+        return self._lookup(NOT)(self) 
         
     def __getitem__(self, indices):
         '''
@@ -558,14 +547,14 @@ class OperatorMixin(object):
                 separator = index
             else:
                 raise TypeError(index)
-        return Global()[REPEAT](self, start, stop, step, separator, add)
+        return self._lookup(REPEAT)(self, start, stop, step, separator, add)
         
     def __gt__(self, function):
         '''
-        **self in function** - Process or label the results.
+        **self > function** - Process or label the results.
         
         Create a named pair or apply a function to the results.  This is
-        equivalent to `lepl.matchers.Apply()`.
+        equivalent to `Apply()`.
         
         :Parameters:
         
@@ -577,18 +566,35 @@ class OperatorMixin(object):
             result.
             
             If a function is given it is called with the results as an
-            argument.  The return value is used as the new result.  This
-            is equivalent to `lepl.matchers.Apply()` with nolist=False.
+            argument.  The return value is used *within a list* as the new 
+            result.  This is equivalent to `Apply()` with raw=False.
         '''
         self.__check(APPLY, function, False)
-        return Global()[APPLY](self, function) 
+        return self._lookup(APPLY)(self, function) 
+    
+    def __ge__(self, function):
+        '''
+        **self >= function** - Process or label the results.
+        
+        Apply a function to the results.  
+        This is equivalent to `Apply(raw=True)`.
+        
+        :Parameters:
+        
+          function
+            This is called with the results as an argument.  The return value 
+            is used as the new result.  This is equivalent to `Apply()` with 
+            raw=True.
+        '''
+        self.__check(APPLY_RAW, function, False)
+        return self._lookup(APPLY_RAW)(self, function) 
     
     def __rshift__(self, function):
         '''
         **self >> function** - Process or label the results (map).
         
         Create a named pair or apply a function to each result in turn.  
-        This is equivalent to `lepl.matchers.Map()`.  It is similar to 
+        This is equivalent to `Map()`.  It is similar to 
         *self >= function*, except that the function is applied to each 
         result in turn.
         
@@ -605,32 +611,14 @@ class OperatorMixin(object):
             The return values are used as the new result.
         '''
         self.__check(MAP, function, False)
-        return Global()[MAP](self, function) 
-        
-    def __mul__(self, function):
-        '''
-        **self * function** - Process the results (\*args).
-        
-        Apply a function to each result in turn.  
-        This is equivalent to `lepl.matchers.Apply()` with ``args=True``.  
-        It is similar to *self > function*, except that the function is 
-        applies to multiple arguments (using Python's ``*args`` behaviour).
-        
-        :Parameters:
-        
-          function
-            A function that is called with the results as arguments.
-            The return values are used as the new result.
-        '''
-        self.__check(ARGS, function, False)
-        return Global()[ARGS](self, function) 
+        return self._lookup(MAP)(self, function) 
         
     def __pow__(self, function):
         '''
         **self \** function** - Process the results (\**kargs).
         
         Apply a function to keyword arguments
-        This is equivalent to `lepl.matchers.KApply()`.
+        This is equivalent to `KApply()`.
         
         :Parameters:
         
@@ -646,14 +634,11 @@ class OperatorMixin(object):
               stream_out
                 The stream returned from the matcher.
                 
-              core
-                The core, if streams are being used, else ``None``.
-            
               results
                 A list of the results returned.
         '''
         self.__check(KARGS, function, False)
-        return Global()[KARGS](self, function) 
+        return self._lookup(KARGS)(self, function) 
     
     def __xor__(self, message):
         '''
@@ -666,7 +651,7 @@ class OperatorMixin(object):
           message
             The message for the SyntaxError.
         '''
-        return Global()[RAISE](self, message)
+        return self._lookup(RAISE)(self, message)
                              
     def __check(self, name, other, is_match):
         '''
